// BSP : STM32 boards groups by STMicroelectronics 2.7.1
// Borad select : STM Boards groups / LoRa boars
// Board part number : LoRa-E5 mini (Board to be selected from Tools submenu 'Board part number)

// 1. wait for transmission of Slave
// 2. receives Slave data(Slave UID, mills(), Vbatt)
// 3. transtits response data(Master UID, rssi, snr, reserved)

// 2024/02/11

#include <Adafruit_GFX.h>       // 1.11.9
#include <Adafruit_SSD1306.h>   // 2.5.9
#include "myFonts/tahoma5pt7b.h"  // any font
#include "myFonts/tahoma7pt7b.h"
#include <RadioLib.h>           // 6.4.2  https://github.com/jgromes/RadioLib

// BAT+_pin() --> 1000k --> A4_pin --> 510k --> D0_pin
#define PIN_VBATT      A4       // battery voltage measurement pin
#define READ_ENABLE    D0       // read PIN_VBATT enable LOW:enable, HIGH:disable

// RM0461 Reference manual Rev 8  37.1.4 IEEE 64-bit unique device ID register (UID64)
#define MASTER_UID  *((uint32_t*)0x1FFF7580)    // use uid64 dor Master device UID
#define SLAVE_UID   0x05370378  // sample Slave device UID

Adafruit_SSD1306 display(128, 64, &Wire, -1);

// \Arduino15\packages\STMicroelectronics\hardware\stm32\2.7.1
// \variants\STM32WLxx\WL54JCI_WL55JCI_WLE4J(8-B-C)I_WLE5J(8-B-C)I\variant_LORA_E5_MINI.h : 173
STM32WLx radio = new STM32WLx_Module();
static const uint32_t rfswitch_pins[] =
                         {PA4,  PA5,  RADIOLIB_NC};
static const Module::RfSwitchMode_t rfswitch_table[] = {
  {STM32WLx::MODE_IDLE,  {LOW,  LOW}},
  {STM32WLx::MODE_RX,    {HIGH, LOW}},
  {STM32WLx::MODE_TX_HP, {LOW, HIGH}},  // for LoRa-E5 mini
//  {STM32WLx::MODE_TX_LP, {HIGH, HIGH}},   // for LoRa-E5-LE mini
  END_OF_MODE_TABLE,
};

#define dataNum   12                // data bytes
union unionData {                   // for data type conversion
  uint32_t  dataBuff32[dataNum/4];
  float     dataBuff32F[dataNum/4];
  int16_t   dataBuff16[dataNum/2];
  uint8_t   dataBuff8[dataNum];
};
union unionData ud;

int rssi;               // signal RSSI [dBm]
int snr;                // signal SN ratio [dB]
uint32_t masteruid = MASTER_UID;    // Master device UID
uint32_t slaveuid;      // Slave device UID
uint16_t sec;           // mS/1000 data [sec]
float Vbatt;            // battery voltage data [V]
char printBuff[4];      // for sprintf()
String  txdata = "";    // transmission data packet string
String  rxdata = "";    // received data packet string
bool rxstatus = false;  // receive status flag

bool receivedFlag = false;     // flag that a packet was received
bool transmittedFlag = false;  // flag that a packet was transmitted

// *******************************************************************************************************
void setup() {
  Serial.begin(115200);
  while(!Serial);
  delay(2000);

  // Pin initialization
  pinMode(LED_RED, OUTPUT);         // built-in LED
  pinMode(READ_ENABLE, OUTPUT);     // read PIN_VBATT enable

  // Display initialization
  display.begin(SSD1306_SWITCHCAPVCC, 0x3C);
  display.clearDisplay();
  display.setFont(&tahoma7pt7b);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 15);
  display.println("LoRa-E5");
  display.setCursor(0, 31);
  display.println("P2P Test");
  display.setCursor(0, 47);
  display.println("Master");  
  display.display();
  delay(2000);
  display.clearDisplay();   


  // STM32WL initialization
  radio.setRfSwitchTable(rfswitch_pins, rfswitch_table);

  // int16_t begin(float freq = 434.0, float bw = 125.0, uint8_t sf = 9, uint8_t cr = 7, \
  // uint8_t syncWord = RADIOLIB_SX126X_SYNC_WORD_PRIVATE, int8_t power = 10, \
  // uint16_t preambleLength = 8, float tcxoVoltage = 1.6, bool useRegulatorLDO = false);
  int state = radio.begin(922.6, 125.0, 12, 5, 0x12, 13, 8, 1.7, 0);

  if (state == RADIOLIB_ERR_NONE) {
    Serial.println("radio.begin() success!");
  }
  else {
    Serial.print("failed, code ");
    Serial.println(state);
    while(true) {
      digitalWrite(LED_RED, LOW);
      delay(100);
      digitalWrite(LED_RED, HIGH);
      delay(100);      
    }
  }

  // callback function when received or transmitted
  radio.setDio1Action(setFlag);

  // setup end 
  display.setCursor(0, 47);
  display.print("LoRa-E5 Master");
  display.setCursor(0, 63);
  display.print("Setup END");
  display.display();
  delay(2000);
  display.clearDisplay(); 
}

// callback function when received or transmitted
void setFlag(void) 
{
  uint16_t irqstatus = radio.getIrqStatus();
  if(irqstatus == RADIOLIB_SX126X_IRQ_RX_DONE) {
    receivedFlag = true;
  }
  else if(irqstatus == RADIOLIB_SX126X_IRQ_TX_DONE) {
    transmittedFlag = true;
  }
  else {
    receivedFlag = false;
    transmittedFlag = false;
  }
}


// ***********************************************************************************************************
void loop() 
{
  // Master receives packet from Slave *************************************************************************
  Serial.println("Waiting for incoming Slave transmitting packet ... ");

  // start listening for LoRa packets
  int state = radio.startReceive();
  if (state != RADIOLIB_ERR_NONE) {
    Serial.print("failed, code ");
    Serial.println(state);
    while (true) {
      digitalWrite(LED_RED, LOW);
      delay(100);
      digitalWrite(LED_RED, HIGH);
      delay(100);    
    }
  }

  // check if the flag is set
  while(!receivedFlag) {
    while(SubGhz.isBusy());
  }
  receivedFlag = false;

  // read a packet 
  state = radio.readData(ud.dataBuff8, dataNum);

  slaveuid = ud.dataBuff32[0];    // Slave device UID
  if (state == RADIOLIB_ERR_NONE & slaveuid == SLAVE_UID) {   
    // packet form Slave was successfully received
    rxdata = "";
    for(int i = 0; i < dataNum; i++) {
      sprintf(printBuff, "%02x", ud.dataBuff8[i]);
      rxdata += printBuff;
    }
    rxdata.toUpperCase();
    Serial.println(rxdata);

    // restoration of received data
    sec = ud.dataBuff32[1] / 1000;      // transmission time
    Vbatt = ud.dataBuff32F[2];          // battery voltage

    Serial.printf("Slave UID:\t%08x\n", slaveuid);
    Serial.print("Time:\t\t"); Serial.print(sec); Serial.println(" sec");
    Serial.print("Vbatt:\t\t"); Serial.print(Vbatt, 3); Serial.println(" V");

    // RSSI and SNR of the last received packet
    Serial.print("RSSI:\t\t");
    rssi = radio.getRSSI();
    Serial.print(rssi);
    Serial.println(" dBm");

    Serial.print("SNR:\t\t");
    snr = radio.getSNR();
    Serial.print(snr);
    Serial.println(" dB");

    display.clearDisplay();
    display.setCursor(0, 15); display.print("from");
    display.setCursor(64, 15); display.print(slaveuid, HEX);  
    display.setFont(&tahoma5pt7b);
    display.setCursor(0, 31); display.print(rxdata);
    display.setFont(&tahoma7pt7b);
    display.setCursor(10, 47); display.print(rssi); display.print(" dBm");
    display.setCursor(64, 47); display.print(snr); display.print(" dB");
    display.setCursor(0, 63); display.print(sec); display.print(" sec");
    display.setCursor(64, 63); display.print(Vbatt, 3); display.print(" V");   
    display.display();
    delay(1000); // for display
  }
  else {
    // some other error occurred
    Serial.print("failed, code ");
    Serial.println(state);
  }  


  // Master transmits response data to Slave  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  Serial.println("Transmitting response packet ... ");

  // send data packet
  ud.dataBuff32[0] = masteruid;     // Master device UID
  ud.dataBuff16[2] = rssi;          // Master RSSI  dBm
  ud.dataBuff16[3] = snr;           // Master SN ratio  dB
  ud.dataBuff32F[2] = 0.0;          // reserved

  digitalWrite(LED_RED, LOW);
  transmittedFlag = false;
  state = radio.startTransmit(ud.dataBuff8, dataNum);
  digitalWrite(LED_RED, HIGH);

  // print the packet
  txdata = "";
  for(int i = 0; i < dataNum; i++) {
    sprintf(printBuff, "%02x", ud.dataBuff8[i]);
    txdata += printBuff;
  }
  txdata.toUpperCase();
  Serial.println(txdata);

  Serial.printf("Master UID:\t%08x\n", masteruid);
  Serial.print("RSSI\t\t"); Serial.print(rssi); Serial.println(" dBm");
  Serial.print("SNR\t\t"); Serial.print(snr); Serial.println(" dB");

  display.clearDisplay();
  display.setCursor(0, 15); display.print("Master");
  display.setCursor(64, 15); display.print(masteruid, HEX);  
  display.setFont(&tahoma5pt7b);
  display.setCursor(0, 31); display.print(txdata);
  display.setFont(&tahoma7pt7b);
  display.setCursor(10, 47); display.print(rssi); display.print(" dBm");
  display.setCursor(64, 47); display.print(snr); display.print(" dB");
  display.setCursor(0, 63); display.print(sec); display.print(" sec");
  display.setCursor(64, 63); display.print(Vbatt, 3); display.print(" V");
  display.display();   
   
  // wait for transmittion completion
  while(!transmittedFlag) {
    while(SubGhz.isBusy());
  }

  // error status check
  if (state == RADIOLIB_ERR_NONE) {
    // packet was successfully sent
    Serial.println("transmission finished!");
    Serial.println();
  } 
  else {
    // some other error occurred
    Serial.print("failed, code ");
    Serial.println(state);
  }
}

